/*
 * AUTHOR:   N.Cihan Tas
 * MODIFIED: Ladan Gharai
 *           Colin Perkins
 * 
 * This program is a linked list for the playout buffer.
 *
 * Copyright (c) 2003-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.48 $
 * $Date: 2004/04/30 18:02:08 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "tv.h"
#include "rtp/rtp.h"
#include "rtp/rtp_callback.h"
#include "rtp/ptime.h"
#include "rtp/pbuf.h"
#include "rtp/decoders.h"

#define PBUF_MAGIC	0xcafebabe

struct pbuf_node {
        struct pbuf_node  *nxt;
        struct pbuf_node  *prv;
        uint32_t           rtp_timestamp;	/* RTP timestamp for the frame           */
        struct timeval     arrival_time;	/* Arrival time of first packet in frame */
        struct timeval     playout_time;	/* Playout time for the frame            */
        struct coded_data *cdata;		/*					 */
	int		   decoded;		/* Non-zero if we've decoded this frame  */
	int     	   mbit;		/* determines if mbit of frame had been seen */
	uint32_t	   magic;		/* For debugging                         */
	int                no;			/* for debugging */
};

struct pbuf {
	struct pbuf_node	*frst;
	struct pbuf_node	*last;
};

/*********************************************************************************/

static void
pbuf_validate(struct pbuf *playout_buf)
{
	/* Run through the entire playout buffer, checking pointers, etc.  */
	/* Only used in debugging mode, since it's a lot of overhead [csp] */
#ifdef NDEF
	struct pbuf_node	*cpb, *ppb;
	struct coded_data	*ccd, *pcd;

	cpb = playout_buf->frst;
	ppb = NULL;
	while (cpb != NULL) {
		assert(cpb->magic == PBUF_MAGIC);
		assert(cpb->prv == ppb);
		if (cpb->prv != NULL) {
			assert(cpb->prv->nxt == cpb);
			/* stored in RTP timestamp order */
			assert(cpb->rtp_timestamp > ppb->rtp_timestamp);	
			/* stored in playout time order  */
			assert(tv_gt(cpb->ptime, ppb->ptime));	
		}
		if (cpb->nxt != NULL) {
			assert(cpb->nxt->prv == cpb);
		} else {
			assert(cpb = playout_buf->last);
		}
		if (cpb->cdata != NULL) {
		/* We have coded data... check all the pointers on that list too */
			ccd = cpb->cdata;
			pcd = NULL;
			while (ccd != NULL) {
				assert(ccd->prv == pcd);
				if (ccd->prv != NULL) {
					assert(ccd->prv->nxt == ccd);
					/* list is descending - cant really check this now*/
					//assert(ccd->seqno < pcd->seqno); 
					assert(ccd->data != NULL);
				}
				if (ccd->nxt != NULL) {
					assert(ccd->nxt->prv == ccd);
				}
				pcd = ccd;
				ccd = ccd->nxt;
			}
		}
		ppb = cpb;
		cpb = cpb->nxt;
	}
#else
	UNUSED(playout_buf);
#endif
}

struct pbuf *
pbuf_init(void)
{  
	struct pbuf	*playout_buf = NULL;

	playout_buf = malloc(sizeof(struct pbuf));
	if (playout_buf != NULL) {
		playout_buf->frst=NULL;
		playout_buf->last=NULL;
	} else {
		debug_msg("Failed to allocate memory for playout buffer\n");
	}
	return playout_buf;
}

static void
add_coded_unit(struct pbuf_node *node, rtp_packet *pkt) 
{
	/* Add "pkt" to the frame represented by "node". The "node" has    */
	/* previously been created, and has some coded data already...     */

	/* New arrivals are added at the head of the list, which is stored */
	/* in descending order of packets as they arrive (NOT necessarily  */
	/* descending sequence number order, as the network might reorder) */

	struct coded_data *tmp; 

	assert(node->rtp_timestamp == pkt->ts);
	assert(node->cdata != NULL);

	tmp = malloc(sizeof(struct coded_data));
	if (tmp != NULL) {
		tmp->seqno  = pkt->seq;	
		tmp->data   = pkt;
		tmp->prv    = NULL;
		tmp->nxt    = node->cdata;
		node->cdata->prv = tmp;
		node->cdata = tmp;
		node->mbit |= pkt->m;
	} else {
		/* this is bad, out of memory, drop the packet... */
		free(pkt);
	}
}

static struct pbuf_node * 
create_new_pnode(rtp_packet *pkt) 
{
	struct pbuf_node *tmp;
	static int no = 0;

	tmp = malloc(sizeof(struct pbuf_node));
	if (tmp != NULL) {
		tmp->magic         = PBUF_MAGIC;
		tmp->nxt           = NULL;
		tmp->prv           = NULL;
		tmp->decoded       = 0;
		tmp->rtp_timestamp = pkt->ts;
		tmp->no		   = no++;
		tmp->mbit	   = pkt->m; 
		gettimeofday(&(tmp->arrival_time), NULL);
		/* FIX ME - need a playout delay */
		gettimeofday(&(tmp->playout_time), NULL);	
	
		tmp->cdata = malloc(sizeof(struct coded_data));
		if (tmp->cdata != NULL) {
			tmp->cdata->nxt   = NULL;
			tmp->cdata->prv   = NULL;
			tmp->cdata->seqno = pkt->seq;
			tmp->cdata->data  = pkt;  /* for now? */
		} else {
			free(pkt); 	
			free(tmp);
			return NULL;
		}
	} else {
		free(pkt);
	}
	return tmp;
}


void
pbuf_insert(struct pbuf *playout_buf, rtp_packet *pkt) 
{
	struct pbuf_node	*tmp;

	pbuf_validate(playout_buf);

	if (playout_buf->frst==NULL && playout_buf->last==NULL) {
		/* playout buffer is empty - add new frame */
		playout_buf->frst = create_new_pnode(pkt);
		playout_buf->last = playout_buf->frst;
		//printf ("\no %3d %ld %ld ", playout_buf.last->no, playout_buf->last->rtp_timestamp, pkt->ts);
		return;
	} 

	if (playout_buf->last->rtp_timestamp == pkt->ts) {
		/* Packet belongs to last frame in playout_buf this is the */
		/* most likely scenario - although...                      */
		add_coded_unit(playout_buf->last, pkt);
	} else {
		if (playout_buf->last->rtp_timestamp < pkt->ts) {
			/* Packet belongs to a new frame... */
			tmp = create_new_pnode (pkt);
			playout_buf->last->nxt = tmp;
			tmp->prv = playout_buf->last;
			playout_buf->last = tmp;
			//printf ("\n  %3d %ld %ld %d", playout_buf->last->no,
			//                       playout_buf->last->rtp_timestamp, pkt->ts, playout_buf->last->prv->m); 
		} else {
			/* Packet belongs to a previous frame... */
			if (playout_buf->frst->rtp_timestamp > pkt->ts) {
				printf("A very old packet - discarded\n");
			} else {
				printf("A packet for a previous frame, but might still be useful\n");
				/* Should probably insert this into the playout buffer here... */
			}
			if (pkt->m) {
				printf("Oops... dropped packet with M bit set\n");
			}
			free(pkt);
		} 
	}
	pbuf_validate(playout_buf);
}

static void
free_cdata (struct coded_data *head)
{
  struct coded_data *tmp;

  while (head) {
	free (head->data);
	tmp = head;
 	head = head->nxt;
	free (tmp);
  }
}

void
pbuf_remove(struct pbuf *playout_buf, struct timeval curr_time)
{
	/* Remove the next frame to be displayed from the playout buffer.  */
	/* The frame should have been previously decoded by pbuf_decode()  */
	/* Returns NULL if there is no frame that has reached its playout  */
	/* time.                                                           */
	/* free the cdata node: the packets should have been free'd before */

	unsigned char *frame = NULL;
	struct pbuf_node *tmp, *frst, *xxx;

	UNUSED(curr_time);
	pbuf_validate(playout_buf);

        tmp=playout_buf->frst;
	/* find a decoded frame */
        while (!(tmp==NULL) && !(frame)) {
		if (tmp->decoded) {
			frst = playout_buf->frst;
			while (!(frst==tmp)) {
				/* these must be incomplete frames */
				xxx = frst;
				//printf("\nd: %d %d", xxx->no, xxx->mbit);
				frst = frst->nxt;
				free_cdata (xxx->cdata);
				free (xxx);
			}
			frst->prv=NULL;
			if (tmp->nxt) {
				playout_buf->frst = tmp->nxt;
			} else {
				/* was last node is pbuf */
				playout_buf->frst=NULL;
				playout_buf->last=NULL;
			}
			free (tmp);
			return;
		}
		tmp=tmp->nxt; 
        }

	pbuf_validate(playout_buf);
	return;
}

static int
frame_complete (struct pbuf_node *frame)
{
        /* Return non-zero if the list of coded_data represents a    */
	/* complete frame of video. This might have to be passed the */ 
	/* seqnum of the last packet in the previous frame, too?     */
	/* i dont think that would reflect correctly of weather this */ 
	/* frame is complete or not - however we should check for all*/
	/* the packtes of a frame being present - perhaps we should  */ 
	/* keep a bit vector in pbuf_node? LG.  */ 
	
	return (frame->mbit==1); 

	/* FIX ME: */ 
	/* adjust pbuf_insert() and add flag to struct pbuf_node */ 
}

int
pbuf_decode(struct pbuf *playout_buf, struct timeval curr_time, unsigned char *framebuffer)
{
	/* This should pass through the playout buffer, and decode one */
	/* complete frame, and exit. The curr_time is passed in as a   */
	/* hint, so that decoding can be delayed until near the time a */
	/* frame is to be used (a future optimization hint, really).   */
	struct pbuf_node	*curr;

	UNUSED(curr_time);

	pbuf_validate(playout_buf);

        curr = playout_buf->frst;
        while (curr != NULL) {
		/* Find a complete and NOT yet decoded frame...  */
		if (curr->decoded) {
			/* Alread decoded... */
			return 0;
		} else if (frame_complete(curr)) {  
			/* decode_frame() must free the coded units hanging off a node's */
			/* cdata, and malloc() space for the decoded_frame.              */
                        decode_frame(curr->cdata, framebuffer);
			curr->cdata   = NULL;
			curr->decoded = 1;
			return 1; 
                }
                curr = curr->nxt;
        }
	return 0;
}

