/*
 * FILE:     transmit.c
 * AUTHOR:   Colin Perkins <csp@csperkins.org>
 * MODIFIED: Ladan Gharai
 *
 * Copyright (c) 2001-2004 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.9 $
 * $Date: 2004/04/12 10:54:20 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "video_types.h"
#include "rtp/rtp.h"
#include "tv.h"
#include "transmit.h"

#define TRANSMIT_MAGIC	0xe80ab15f

struct video_tx {
	uint32_t	 magic;
	unsigned	 mtu;
};

struct video_tx *
tx_init(unsigned mtu)
{
	struct video_tx	*tx;

	tx = (struct video_tx *) malloc(sizeof(struct video_tx));
	if (tx != NULL) {
		tx->magic = TRANSMIT_MAGIC;
		tx->mtu   = mtu;
	}
	return tx;
}

void
tx_done(struct video_tx *tx)
{
	assert(tx->magic == TRANSMIT_MAGIC);
	free(tx);
}

#define HD_WIDTH 1280
#define HD_HEIGHT 720
#define HD_DEPTH    2
#define TS_WRAP     4294967296

typedef struct {
	uint16_t	scan_line;	/* pixels */
	uint16_t	scan_offset;	/* pixels */
	uint16_t	length;		/* octets */
	uint16_t	flags;
} payload_hdr_t;

void
tx_send(struct video_tx *tx, struct video_frame *frame, struct rtp *rtp_session)
{
	int		 m, x, y, first_x, first_y, data_len, l, payload_count, octets_left_this_line, octets_left_this_packet;
	payload_hdr_t	 payload_hdr[10];
	int		 pt = 96;	/* A dynamic payload type for the tests... */
	static uint32_t	 ts = 0;
	char		*data;

	assert(tx->magic == TRANSMIT_MAGIC);

	assert(frame->data_len == (1280 * 720 * 2));
	data = frame->data + (1280 * 720 * 2);

	m = 0;
	x = 0; 
	y = 0;
	payload_count = 0;
	data_len = 0;
	first_x  = x;
	first_y  = y;
	ts = get_local_mediatime();
	do {
		if (payload_count == 0) {
			data_len = 0;
			first_x  = x;
			first_y  = y;
		}

		octets_left_this_line   = (HD_WIDTH - x) * HD_DEPTH;
		octets_left_this_packet = tx->mtu - 40 - data_len - (8 * (payload_count + 1));
		if (octets_left_this_packet < octets_left_this_line) {
			l = octets_left_this_packet;
		} else {
			l = octets_left_this_line;
		}
		while ((l % HD_DEPTH) != 0) {	/* Only send complete pixels */
			l--;
		}

		payload_hdr[payload_count].scan_line   = htons(y);
		payload_hdr[payload_count].scan_offset = htons(x);
		payload_hdr[payload_count].length      = htons(l);
		payload_hdr[payload_count].flags       = htons(0);
		payload_count++;

		data_len = data_len + l;

		x += (l / HD_DEPTH);
		if (x == HD_WIDTH) {
			x = 0;
			y++;
		}
		if (y == HD_HEIGHT) {
			m = 1;
		}

		/* Is it time to send this packet? */
		if ((y == HD_HEIGHT) || (payload_count == 10) || ((40u + data_len + (8u * (payload_count + 1)) + HD_DEPTH) > tx->mtu)) {
			payload_hdr[payload_count - 1].flags = htons(1<<15);
			data = frame->data + (first_y * HD_WIDTH * HD_DEPTH) + (first_x * HD_DEPTH);
			rtp_send_data_hdr(rtp_session, ts, pt, m, 0, 0, (char *) payload_hdr, 8 * payload_count, data, data_len, 0, 0, 0);
			payload_count = 0;
		}
	} while (y < HD_HEIGHT);
}

