/*
 * FILE:   vidcap.c
 * AUTHOR: Colin Perkins <csp@isi.edu>
 *
 * Copyright (c) 2001-2003 University of Southern California
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, is permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 
 *      This product includes software developed by the University of Southern
 *      California Information Sciences Institute.
 * 
 * 4. Neither the name of the University nor of the Institute may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Revision: 1.2 $
 * $Date: 2004/04/30 10:46:10 $
 *
 */

#include "config.h"
#include "config_unix.h"
#include "config_win32.h"
#include "debug.h"
#include "video_types.h"
#include "video_capture.h"
#include "video_capture/hdstation.h"
#include "video_capture/testcard.h"
#include "video_capture/null.h"

#define VIDCAP_MAGIC	0x76ae98f0

struct vidcap {
	void			*state;
	int			 index;
	uint32_t		 magic;		/* For debugging */
};

struct vidcap_device_api {
	vidcap_id_t		 id;
	struct vidcap_type	*(*func_probe)(void);
	void			*(*func_init)(int fps);
	void			 (*func_done)(void *state);
	struct video_frame	*(*func_grab)(void *state);
};

struct vidcap_device_api vidcap_device_table[] = {
#ifdef HAVE_HDSTATION
	{
		/* The DVS HDstation capture card */
		0,
		vidcap_hdstation_probe,
		vidcap_hdstation_init,
		vidcap_hdstation_done,
		vidcap_hdstation_grab
	},
#endif
	{
		/* Dummy sender for testing purposes */
		0,
		vidcap_testcard_probe,
		vidcap_testcard_init,
		vidcap_testcard_done,
		vidcap_testcard_grab
	},
	{
		0,
		vidcap_null_probe,
		vidcap_null_init,
		vidcap_null_done,
		vidcap_null_grab
	}
};

#define VIDCAP_DEVICE_TABLE_SIZE (sizeof(vidcap_device_table)/sizeof(struct vidcap_device_api))

/* API for probing capture devices ****************************************************************/

static struct vidcap_type	*available_devices[VIDCAP_DEVICE_TABLE_SIZE];
static int			 available_device_count = 0;

int
vidcap_init_devices(void)
{
	unsigned int		 i;
	struct vidcap_type	*dt;

	assert(available_device_count == 0);

	for (i = 0; i < VIDCAP_DEVICE_TABLE_SIZE; i++) {
		dt = vidcap_device_table[i].func_probe();
		if (dt != NULL) {
			vidcap_device_table[i].id = dt->id;
			available_devices[available_device_count++] = dt;
		}
	}
	return 0;
}

void
vidcap_free_devices(void)
{
	int	i;

	for (i = 0; i < available_device_count; i++) {
		free(available_devices[i]);
		available_devices[i] = NULL;
	}
	available_device_count = 0;
}

int
vidcap_get_device_count(void)
{
	return available_device_count;
}

struct vidcap_type *
vidcap_get_device_details(int index)
{
	assert(index < available_device_count);
	assert(available_devices[index] != NULL);

	return available_devices[index];
}

vidcap_id_t
vidcap_get_null_device_id(void)
{
	return VIDCAP_NULL_ID;
}

/* API for video capture **************************************************************************/

struct vidcap *
vidcap_init(vidcap_id_t id, int fps)
{
	unsigned int	i;

	for (i = 0; i < VIDCAP_DEVICE_TABLE_SIZE; i++) {
		if (vidcap_device_table[i].id == id) {
			struct vidcap *d = (struct vidcap *) malloc(sizeof(struct vidcap));
			d->magic = VIDCAP_MAGIC;
			d->state = vidcap_device_table[i].func_init(fps);
			d->index = i;
			if (d->state == NULL) {
				debug_msg("Unable to start video capture device 0x%08lx\n", id);
				free(d);
				return NULL;
			} 
			return d;
		}
	}
	debug_msg("Unknown video capture device: 0x%08x\n", id);
	return NULL;
}

void 
vidcap_done(struct vidcap *state)
{
	assert(state->magic == VIDCAP_MAGIC);
	vidcap_device_table[state->index].func_done(state->state);
	free(state);
}

struct video_frame *
vidcap_grab(struct vidcap *state)
{
	assert(state->magic == VIDCAP_MAGIC);
	return vidcap_device_table[state->index].func_grab(state->state);
}

